/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ExampleMemoryClient
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ExampleMemoryClient.h"

#include <RobotAPI/components/armem/server/ExampleMemory/aron/ExampleData.aron.generated.h>

#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/aron_conversions.h>
#include <RobotAPI/libraries/armem/core/operations.h>
#include <RobotAPI/libraries/armem/core/wm/ice_conversions.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

#include <SimoxUtility/color/cmaps.h>
#include <SimoxUtility/math/pose/pose.h>

#include <Eigen/Geometry>

#include <random>


#define STORE_IMAGES 0

#if STORE_IMAGES
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#endif


namespace armarx
{

    armarx::PropertyDefinitionsPtr ExampleMemoryClient::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(p.usedMemoryName, "mem.UsedMemoryName", "Name of the memory to use.");
        defs->optional(p.commitFrequency, "ex.CommitFrequency", "Frequency in which example data is commited.");

        return defs;
    }


    std::string ExampleMemoryClient::getDefaultName() const
    {
        return "ExampleMemoryClient";
    }


    void ExampleMemoryClient::onInitComponent()
    {
    }


    void ExampleMemoryClient::onConnectComponent()
    {
        createRemoteGuiTab();
        RemoteGui_startRunningTask();

        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "Waiting for memory '" << p.usedMemoryName << "' ...";
        try
        {
            memoryReader = memoryNameSystem().useReader(p.usedMemoryName);
            memoryWriter = memoryNameSystem().useWriter(p.usedMemoryName);
        }
        catch (const armem::error::CouldNotResolveMemoryServer& e)
        {
            ARMARX_ERROR << e.what();
            return;
        }

        // Add a provider segment to commit to.
        exampleProviderID = addProviderSegment();
        // Construct the entity ID.
        exampleEntityID = exampleProviderID.withEntityName("example_entity");


        // Subscribe to example_entity updates
        // Using a lambda:
        memoryNameSystem().subscribe(
            exampleEntityID,
            [&](const armem::MemoryID & exampleEntityID, const std::vector<armem::MemoryID>& snapshotIDs)
        {
            ARMARX_INFO << "Entity " << exampleEntityID << " was updated by " << snapshotIDs.size() << " snapshots.";
        });
        // Using a member function:
        memoryNameSystem().subscribe(exampleEntityID, this, &ExampleMemoryClient::processExampleEntityUpdate);


        task = new RunningTask<ExampleMemoryClient>(this, &ExampleMemoryClient::run);
        task->start();
    }


    void ExampleMemoryClient::onDisconnectComponent()
    {
        task->stop();
    }


    void ExampleMemoryClient::onExitComponent()
    {
    }


    void ExampleMemoryClient::run()
    {
        ARMARX_IMPORTANT << "Running example.";
        run_started = IceUtil::Time::now();

        armem::MemoryID snapshotID = commitSingleSnapshot(exampleEntityID);
        if (true)
        {
            commitMultipleSnapshots(exampleEntityID, 3);
        }
        if (true)
        {
            queryLatestSnapshot(snapshotID.getEntityID());
        }
        if (true)
        {
            queryExactSnapshot(snapshotID);
        }
        if (true)
        {
            commitExampleData();
            queryExampleData();
        }
        if (true)
        {
            commitExamplesWithLinks();
        }

        CycleUtil c(static_cast<int>(1000 / p.commitFrequency));
        int i = 0;
        while (!task->isStopped())
        {
            commitExampleData();
            if (i++ < 10000)
            {
                commitSingleSnapshot(exampleEntityID);
            }
            c.waitForCycleDuration();
        }
    }


    armem::MemoryID ExampleMemoryClient::addProviderSegment()
    {
        armem::data::AddSegmentInput input;
        input.coreSegmentName = "ExampleModality";
        input.providerSegmentName = "FancyMethodModality";

        ARMARX_IMPORTANT << input;
        armem::data::AddSegmentResult result = memoryWriter.addSegment(input);
        ARMARX_INFO << result;

        return armem::MemoryID(result.segmentID);
    }


    armem::MemoryID ExampleMemoryClient::commitSingleSnapshot(const armem::MemoryID& entityID)
    {
        std::default_random_engine gen(std::random_device{}());
        std::uniform_int_distribution<int> distrib(-20, 20);

        // Prepare the update with some empty instances.
        armem::EntityUpdate update;
        update.entityID = entityID;
        update.timeCreated = armem::Time::now();

        double diff = (update.timeCreated - run_started).toMilliSecondsDouble() / 1000;

        auto dict1 = std::make_shared<aron::datanavigator::DictNavigator>();
        auto dict2 = std::make_shared<aron::datanavigator::DictNavigator>();

        auto sin = std::make_shared<aron::datanavigator::FloatNavigator>(std::sin(diff));
        auto cos = std::make_shared<aron::datanavigator::FloatNavigator>(std::cos(diff));

        auto sqrt = std::make_shared<aron::datanavigator::DoubleNavigator>(std::sqrt(diff));
        auto lin = std::make_shared<aron::datanavigator::LongNavigator>(static_cast<long>(diff * 1000));
        auto rand = std::make_shared<aron::datanavigator::IntNavigator>(distrib(gen));

        dict1->addElement("sin", sin);
        dict1->addElement("cos", cos);

        dict2->addElement("sqrt", sqrt);
        dict2->addElement("lin", lin);
        dict2->addElement("rand", rand);

        update.instancesData =
        {
            dict1,
            dict2
        };

        ARMARX_IMPORTANT << "Committing " << update;
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);
        ARMARX_INFO << updateResult;
        if (!updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
        }

        return updateResult.snapshotID;
    }


    void ExampleMemoryClient::commitMultipleSnapshots(const armem::MemoryID& entityID, int num)
    {
        // Commit a number of updates with different timestamps and number of instances.
        armem::Commit commit;
        for (int i = 0; i < num; ++i)
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = entityID;
            update.timeCreated = armem::Time::now() + armem::Time::seconds(i);
            for (int j = 0; j < i; ++j)
            {
                update.instancesData.push_back(std::make_shared<aron::datanavigator::DictNavigator>());
            }
        }
        ARMARX_IMPORTANT << "Committing " << commit;
        armem::CommitResult commitResult = memoryWriter.commit(commit);
        ARMARX_INFO << commitResult;
        if (!commitResult.allSuccess())
        {
            ARMARX_ERROR << commitResult.allErrorMessages();
        }
    }


    void ExampleMemoryClient::queryLatestSnapshot(const armem::MemoryID& entityID)
    {
        ARMARX_IMPORTANT
                << "Querying latest snapshot: "
                << "\n- entityID:     \t'" << entityID << "'"
                ;

        armem::client::query::Builder builder;
        builder
        .coreSegments().withID(entityID)
        .providerSegments().withID(entityID)
        .entities().withID(entityID)
        .snapshots().latest();

        armem::client::QueryResult qResult = memoryReader.query(builder.buildQueryInput());
        ARMARX_INFO << qResult;
        if (qResult.success)
        {
            ARMARX_IMPORTANT << "Getting entity via ID";

            armem::wm::Memory& memory = qResult.memory;
            ARMARX_CHECK_GREATER_EQUAL(memory.size(), 1);

            const armem::wm::Entity* entity = memory.findEntity(entityID);
            ARMARX_CHECK_NOT_NULL(entity)
                    << "Entity " << entityID << " was not found in " << armem::print(memory);
            ARMARX_CHECK_GREATER_EQUAL(entity->size(), 1);

            const armem::wm::EntitySnapshot& snapshot = entity->getLatestSnapshot();
            ARMARX_CHECK_GREATER_EQUAL(snapshot.size(), 1);

            ARMARX_INFO << "Result: "
                        << "\n- entity:       \t" << entity->name()
                        << "\n- snapshot:     \t" << snapshot.time()
                        << "\n- #instances:   \t" << snapshot.size()
                        ;

            // Show memory contents in remote gui.
            tab.queryResult = std::move(memory);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }
    }


    void ExampleMemoryClient::queryExactSnapshot(const armem::MemoryID& snapshotID)
    {
        ARMARX_IMPORTANT
                << "Querying exact snapshot: "
                << "\n- snapshotID:     \t'" << snapshotID << "'"
                ;

        namespace qf = armem::client::query_fns;
        armem::client::query::Builder qb;
        qb.singleEntitySnapshot(snapshotID);

        armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());
        ARMARX_INFO << qResult;

        if (qResult.success)
        {
            armem::wm::Memory memory = std::move(qResult.memory);
            const armem::wm::EntitySnapshot& entitySnapshot = memory.getEntity(snapshotID).getLatestSnapshot();

            ARMARX_INFO << "Result snapshot: "
                        << "\n- time:        \t" << entitySnapshot.time()
                        << "\n- # instances: \t" << entitySnapshot.size()
                        ;
        }
        else
        {
            ARMARX_ERROR << qResult.errorMessage;
        }
    }


    void ExampleMemoryClient::commitExampleData()
    {
        ARMARX_IMPORTANT << "Adding segment " << "ExampleData" << "/" << getName();

        auto addSegmentResult = memoryWriter.addSegment("ExampleData", getName());
        if (!addSegmentResult.success)
        {
            ARMARX_ERROR << addSegmentResult.errorMessage;
            return;
        }
        exampleDataProviderID = armem::MemoryID(addSegmentResult.segmentID);

        const armem::Time time = armem::Time::now();
        armem::Commit commit;

        //default
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("default");
            update.timeCreated = time;

            armem::example::ExampleData data;
            toAron(data.memoryLink, armem::MemoryID());
            update.instancesData = { data.toAron() };
        }
        //the answer
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("the answer");
            update.timeCreated = time;

            armem::example::ExampleData data;
            data.the_bool = true;
            data.the_double = std::sin(time.toSecondsDouble());
            data.the_float = 21.5;
            data.the_int = 42;
            data.the_long = 424242;
            data.the_string = "fourty two";
            data.the_float_list = { 21, 42, 84 };
            data.the_int_list = { 21, 42, 84 };
            data.the_string_list = simox::alg::multi_to_string(data.the_int_list);
            data.the_object_list.emplace_back();

            data.the_float_dict =
            {
                { "one", 1.0 },
                { "two", 2.0 },
                { "three", 3.0 },
            };
            data.the_int_dict =
            {
                { "one", 1 },
                { "two", 2 },
                { "three", 3 },
            };

            data.the_position = { 42, 24, 4224 };
            data.the_orientation = Eigen::AngleAxisf(1.57f, Eigen::Vector3f(1, 1, 1).normalized());
            data.the_pose = simox::math::pose(data.the_position, data.the_orientation);

            data.the_3x1_vector = { 24, 42, 2442 };
            data.the_4x4_matrix = 42 * Eigen::Matrix4f::Identity();

            toAron(data.memoryLink, armem::MemoryID());

            simox::ColorMap cmap = simox::color::cmaps::plasma();
            {
                cv::Mat& image = data.the_rgb24_image;
                image.create(10, 20, image.type());
                cmap.set_vlimits(0, float(image.cols + image.rows));
                using Pixel = cv::Point3_<uint8_t>;
                image.forEach<Pixel>([&cmap](Pixel& pixel, const int index[]) -> void
                {
                    simox::Color color = cmap(float(index[0] + index[1]));
                    pixel.x = color.r;
                    pixel.y = color.g;
                    pixel.z = color.b;
                });

#if STORE_IMAGES
                cv::Mat out;
                cv::cvtColor(image, out, CV_RGB2BGR);
                cv::imwrite("the_rgb24_image.png", out);
#endif
            }
            {
                cv::Mat& image = data.the_depth32_image;
                image.create(20, 10, image.type());
                image.forEach<float>([&image](float& pixel, const int index[]) -> void
                {
                    pixel = 100 * float(index[0] + index[1]) / float(image.rows + image.cols);
                });

#if STORE_IMAGES
                cmap.set_vlimits(0, 100);
                cv::Mat rgb(image.rows, image.cols, CV_8UC3);
                using Pixel = cv::Point3_<uint8_t>;
                rgb.forEach<Pixel>([&image, &cmap](Pixel& pixel, const int index[]) -> void
                {
                    simox::Color color = cmap(image.at<float>(index));
                    pixel.x = color.r;
                    pixel.y = color.g;
                    pixel.z = color.b;
                });

                cv::Mat out;
                cv::cvtColor(rgb, out, CV_RGB2BGR);
                cv::imwrite("the_depth32_image.png", out);
#endif
            }

            update.instancesData = { data.toAron() };
        }

        armem::CommitResult commitResult = memoryWriter.commit(commit);
        if (commitResult.allSuccess())
        {
            armem::fromIce(commitResult.results.at(1).snapshotID, theAnswerSnapshotID);
        }
        else
        {
            ARMARX_WARNING << commitResult.allErrorMessages();
        }
    }


    void ExampleMemoryClient::queryExampleData()
    {
        // Query all entities from provider.
        armem::client::query::Builder qb;
        qb
        .coreSegments().withID(exampleProviderID)
        .providerSegments().withID(exampleProviderID)
        .entities().all()
        .snapshots().all();

        armem::client::QueryResult result = memoryReader.query(qb.buildQueryInput());
        if (result.success)
        {
            tab.queryResult = std::move(result.memory);
            tab.rebuild = true;
        }
        else
        {
            ARMARX_ERROR << result.errorMessage;
        }
    }


    void ExampleMemoryClient::commitExamplesWithLinks()
    {
        const armem::Time time = armem::Time::now();

        armem::Commit commit;
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("link to the_answer");
            update.timeCreated = time;

            armem::example::ExampleData data;
            armem::toAron(data.memoryLink, theAnswerSnapshotID);

            update.instancesData = { data.toAron() };
        }
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("link to self");
            update.timeCreated = time;

            armem::example::ExampleData data;
            armem::toAron(data.memoryLink, update.entityID.withTimestamp(time));

            update.instancesData = { data.toAron() };
        }

        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("link to previous snapshot");
            update.timeCreated = time - armem::Time::seconds(1);  // 1 sec in the past

            armem::example::ExampleData data;
            armem::toAron(data.memoryLink, armem::MemoryID());  // First entry - invalid link

            update.instancesData = { data.toAron() };
        }
        {
            armem::EntityUpdate& update = commit.add();
            update.entityID = exampleDataProviderID.withEntityName("link to previous snapshot");
            update.timeCreated = time;

            armem::example::ExampleData data;
            armem::toAron(data.memoryLink, update.entityID.withTimestamp(time - armem::Time::seconds(1)));

            update.instancesData = { data.toAron() };
        }

        ARMARX_CHECK_EQUAL(commit.updates.size(), 4);
        armem::CommitResult commitResult = memoryWriter.commit(commit);

        if (!commitResult.allSuccess() || commitResult.results.size() != commit.updates.size())
        {
            ARMARX_WARNING << commitResult.allErrorMessages();
        }
    }


    void ExampleMemoryClient::processExampleEntityUpdate(
        const armem::MemoryID& subscriptionID, const std::vector<armem::MemoryID>& snapshotIDs)
    {
        std::stringstream ss;
        ss << "example_entity got updated: " << subscriptionID << "\n";
        ss << "Updated snapshots: \n";
        for (const auto& id : snapshotIDs)
        {
            ss << "- " << id << "\n";
        }
        ARMARX_IMPORTANT << ss.str();
        // Fetch new data of example_entity and do something with it.
    }


    void ExampleMemoryClient::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        if (tab.queryResult)
        {
            tab.queryResultGroup = armem::server::MemoryRemoteGui().makeGroupBox(*tab.queryResult);
        }

        VBoxLayout root = {tab.queryResultGroup, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }

    void ExampleMemoryClient::RemoteGui_update()
    {
        if (tab.rebuild.exchange(false))
        {
            createRemoteGuiTab();
        }
    }

}
