/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Rainer Kartmann (rainer dot kartmann at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Image.h"

#include <RobotAPI/libraries/aron/core/Exception.h>

#include <SimoxUtility/algorithm/string.h>
#include <SimoxUtility/meta/EnumNames.hpp>


namespace armarx::aron::typenavigator
{

    static const simox::meta::EnumNames<ImagePixelType> ImagePixelTypeNames
    {
        { ImagePixelType::Rgb24, "rgb24" },
        { ImagePixelType::Depth32, "depth32" },
    };


    // constructors
    ImageNavigator::ImageNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eImage, path)
    {
    }


    ImageNavigator::ImageNavigator(const type::AronImagePtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eImage, path),
        detail::NavigatorBase<type::AronImage, ImageNavigator>(o)
    {
    }


    type::AronImagePtr ImageNavigator::toAronImagePtr() const
    {
        return this->aron;
    }


    ImagePixelType ImageNavigator::getPixelType() const
    {
        return pixelTypeFromName(this->aron->pixelType);
    }

    void ImageNavigator::setPixelType(const ImagePixelType& pixelType) const
    {
        this->aron->pixelType = pixelTypeToName(pixelType);
    }

    void ImageNavigator::setPixelType(const std::string& typeName) const
    {
        // Try a lookup, which throws when failing.
        pixelTypeFromName(typeName);
        this->aron->pixelType = typeName;
    }


    // virtual implementations
    std::string ImageNavigator::getName() const
    {
        return "AronImageType";
    }


    bool ImageNavigator::operator==(const ImageNavigator& other) const
    {
        return *this->aron == *other.aron;
    }


    std::string ImageNavigator::pixelTypeToName(ImagePixelType type)
    {
        try
        {
            return ImagePixelTypeNames.to_name(type);
        }
        catch (const simox::meta::error::UnknownEnumValue& e)
        {
            throw aron::error::AronException("ImageNavigator", __FUNCTION__, e.what());
        }
    }


    ImagePixelType ImageNavigator::pixelTypeFromName(const std::string& name)
    {
        try
        {
            return ImagePixelTypeNames.from_name(simox::alg::to_lower(name));
        }
        catch (const simox::meta::error::UnknownEnumValue& e)
        {
            throw aron::error::AronException("ImageNavigator", __FUNCTION__, e.what());
        }
    }


    std::set<ImagePixelType> ImageNavigator::pixelTypes()
    {
        return ImagePixelTypeNames.values();
    }

    std::set<std::string> ImageNavigator::pixelTypeNames()
    {
        return ImagePixelTypeNames.names();
    }

}

